/*
 * SnFTableModel
 *
 * Copyright (c) 2004 ValueCommerce Co.,Ltd. All Rights Reserved
 */
package ru.skipy.ui.components.sft;

import java.util.*;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;

import ru.skipy.filtering.Filter;

/**
 * This implementation of <code>javax.swing.table.TableModel</code> represents filterable
 * and sortable table model.<br>
 * <br>
 * Filters are external and can be set and dropped by
 * {@link #setFilter(ru.skipy.filtering.Filter, boolean)} and {@link #dropFilter()}.<br>
 * <br>
 * Sorting implementation is closed. Only sort preferences can be specified by
 * {@link #setSortPreferences(ColumnSortPrefs[], boolean)}. Sort can be dropped by
 * {@link #dropSortPreferences()}
 *
 * @author Eugene Matyushkin
 * @version 1.0
 */
public class SnFTableModel implements TableModel {

    /**
     * Column classes
     */
    private Class[] columnClasses;

    /**
     * Column names
     */
    private String[] columnNames;

    /**
     * Table data
     */
    private final List<TableRow> tableData;

    /**
     * Filtered and sorted table data
     */
    private final List<TableRow> displayableTableData;

    /**
     * Table filter
     */
    private Filter<TableRow> filter = DumbFilter.INSTANCE;

    /**
     * Table sorters
     */
    private final TableRowSorter sorter;

    /**
     * Table model listeners list
     */
    private final List<TableModelListener> listeners;


    /**
     * Constructs new table model.<br>
     * <br>
     * <b>Notes:</b>
     * <ul>
     * <li> arrays passed shouldn't be null
     * <li> columnClasses size and columnNames size should be equal
     * <li> columnClasses size and columnNames size should NOT be equal to zero
     * <li> each row length in data array should be equal to columnClasses array size
     * <li> each object in row should have class <b>equal</b> to the corresponding
     * class in columnClasses array
     * </ul>
     *
     * @param columnClasses column classes array
     * @param columnNames   column names array
     * @param data          table data
     */
    public SnFTableModel(Class[] columnClasses, String[] columnNames, Object[][] data) {
        if (columnClasses == null || columnClasses.length == 0) {
            throw new IllegalArgumentException("Invalid column classes array!");
        }
        if (columnNames == null || columnNames.length == 0) {
            throw new IllegalArgumentException("Invalid column names array!");
        }
        if (data == null) {
            throw new IllegalArgumentException("Invalid data array!");
        }
        if (columnClasses.length != columnNames.length) {
            throw new IllegalArgumentException("Columns classes number (" + columnClasses.length +
                ") doesn't match column names number (" +
                columnNames.length + ")!");
        }
        this.columnClasses = new Class[columnClasses.length];
        System.arraycopy(columnClasses, 0, this.columnClasses, 0, columnClasses.length);
        this.columnNames = new String[columnNames.length];
        System.arraycopy(columnNames, 0, this.columnNames, 0, columnNames.length);
        TableRowBuilder builder = TableRowBuilderFactory.getInstance(columnClasses);
        tableData = new LinkedList<TableRow>();
        for (Object[] rowData : data) {
            tableData.add(builder.createRow(rowData));
        }
        displayableTableData = new ArrayList<TableRow>(tableData.size());
        sorter = new TableRowSorter(this);
        listeners = new LinkedList<TableModelListener>();
        applyFilter();
    }


    /**
     * Returns row count.<br>
     * <br>
     * <b>Note: </b> row count depends on filter set!
     *
     * @return row count
     */
    public int getRowCount() {
        synchronized (displayableTableData) {
            return displayableTableData.size();
        }
    }


    /**
     * Returns column count
     *
     * @return column count
     */
    public int getColumnCount() {
        return columnClasses.length;
    }


    /**
     * Returns column name.<br>
     * <br>
     * <b>Note: </b> index check is performed!
     *
     * @param columnIndex column index
     *
     * @return column name
     *
     * @throws IllegalArgumentException if index is invalid
     */
    public String getColumnName(int columnIndex) {
        assertColumnIndexIsValid(columnIndex);
        return columnNames[columnIndex];
    }


    /**
     * Returns column class.<br>
     * <br>
     * <b>Note: </b> index check is performed!
     *
     * @param columnIndex column index
     *
     * @return column class
     *
     * @throws IllegalArgumentException if index is invalid
     */
    public Class<?> getColumnClass(int columnIndex) {
        assertColumnIndexIsValid(columnIndex);
        return columnClasses[columnIndex];
    }


    /**
     * Model is not editable by default. Method always return <code>false</code>.<br>
     * <br>
     * <b>Note: </b> indexes' check is performed!
     *
     * @param rowIndex    row index
     * @param columnIndex column index
     *
     * @throws IllegalArgumentException if indexes are out of bounds
     */
    public boolean isCellEditable(int rowIndex, int columnIndex) {
        assertColumnIndexIsValid(columnIndex);
        assertRowIndexIsValid(rowIndex);
        return false;
    }


    /**
     * Returns object by row and column indexes. Indexes are checked for being
     * correct.
     *
     * @param rowIndex    row index
     * @param columnIndex column index
     *
     * @return table data
     *
     * @throws IllegalArgumentException if indexes are invalid
     */
    public Object getValueAt(int rowIndex, int columnIndex) {
        assertColumnIndexIsValid(columnIndex);
        assertRowIndexIsValid(rowIndex);
        synchronized (displayableTableData) {
            return displayableTableData.get(rowIndex).getObjectAt(columnIndex);
        }
    }


    /**
     * Model is not editable by default. Method does not set a value.<br>
     * <br>
     * <b>Note: </b> object class and indexes checks are performed!
     *
     * @param aValue      object to set
     * @param rowIndex    row index
     * @param columnIndex column index
     *
     * @throws IllegalArgumentException if indexes are out of bounds or object's class
     *                                  is invalid
     */
    public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
        assertColumnIndexIsValid(columnIndex);
        assertRowIndexIsValid(rowIndex);
        if (!columnClasses[columnIndex].equals(aValue.getClass())) {
            throw new IllegalArgumentException("Error while trying to set object! Class is invalid: " +
                aValue.getClass().getName() + ". Should be " + columnClasses[columnIndex]);
        }
    }


    /**
     * Adds table model listener
     */
    public void addTableModelListener(TableModelListener listener) {
        listeners.add(listener);
    }


    /**
     * Removes table model listener
     */
    public void removeTableModelListener(TableModelListener listener) {
        listeners.remove(listener);
    }


    /**
     * Checks if column index is in range [0,columnNumber)
     *
     * @param columnIndex row index
     *
     * @throws IllegalArgumentException if column index out of bounds
     */
    private void assertColumnIndexIsValid(int columnIndex) {
        if (columnIndex < 0 || columnIndex >= columnNames.length) {
            throw new IllegalArgumentException(
                "Column index is invalid: " + columnIndex + ". Should be in [0," + columnNames.length + "] range");
        }
    }


    /**
     * Checks if row index is in range [0,displayableRowsNumber)
     *
     * @param rowIndex row index
     *
     * @throws IllegalArgumentException if row index out of bounds
     */
    private void assertRowIndexIsValid(int rowIndex) {
        synchronized (displayableTableData) {
            if (rowIndex < 0 || rowIndex >= displayableTableData.size()) {
                throw new IllegalArgumentException(
                    "Row index is invalid: " + rowIndex + ". Should be in [0," + displayableTableData.size() +
                    "] range");
            }
        }
    }


    /**
     * Sets sort preferences. Table is sorted immediately.
     *
     * @param prefs           sort preferences
     * @param sortImmediately if <code>true</code>, table is sorted immediately
     *
     * @throws IllegalArgumentException if argument is <code>null</code>
     * @throws IllegalArgumentException if some column's index is invalid
     * @throws IllegalArgumentException if any of columns specified is not sortable
     * @see ColumnSortPrefs
     */
    public void setSortPreferences(ColumnSortPrefs[] prefs, boolean sortImmediately) {
        sorter.setSortPrefs(prefs);
        sortTable();
    }


    /**
     * Drops sort preferences. Table is sorted immediately. Initial order is restored.
     */
    public void dropSortPreferences() {
        sorter.setSortPrefs(new ColumnSortPrefs[0]);
        applyFilter();
    }


    /**
     * Sorts table using current settings of sorters. Notifies listeners
     * about table change.
     */
    private void sortTable() {
        synchronized (displayableTableData) {
            Collections.sort(displayableTableData, sorter);
        }
        fireTableChangedEvent(new TableModelEvent(this));
    }


    /**
     * Checks if table can be sorted by specified column. If column class is
     * not a descendant of <code>java.lang.Comparable</code>, comparator can be
     * registered to provide capability of sorting: <br></br>
     * <code>registerComparatorForClass(clazz, comparator)</code><br>
     * <br>
     * <b>Note: </b> index check is performed.
     *
     * @param columnIndex column index
     *
     * @return <code>true</code> if table can be sorted by the column specified
     */
    public boolean canBeSortedByColumn(int columnIndex) {
        assertColumnIndexIsValid(columnIndex);
        Class clazz = columnClasses[columnIndex];
        return (Comparable.class.isAssignableFrom(clazz) || sorter.isComparatorRegistered(clazz));
    }


    /**
     * Registers comparator for the class specified.<br>
     * <br>
     * <b>Note: </b> comparator should be parametrized by specified class. I.e., custom
     * comparator for <code>String</code> (which class is <code>String.class</code> or
     * <code>Class&lt;String&gt;</code>) should be declared as <code><b>impements</b>
     * Comparator&lt;String&gt;</code>
     *
     * @param clazz      class for which comparator is registered
     * @param comparator comparator to be registered
     */
    public synchronized <T> void registerComparatorForClass(Class<T> clazz, Comparator<T> comparator) {
        sorter.registerComparatorForClass(clazz, comparator);
    }


    /**
     * Unregisters comparator for the class specified.
     *
     * @param clazz class, which comparator should be unregistered
     */
    public synchronized void unregisterComparatorForClass(Class clazz) {
        sorter.unregisterComparatorForClass(clazz);
    }


    /**
     * Applies filter. Filtered content will be created. Initial content should
     * not be lost.<br>
     * <br>
     * <b>Note: </b> table is sorted after filtering using current settings
     */
    private void applyFilter() {
        synchronized (displayableTableData) {
            displayableTableData.clear();
            for (TableRow row : tableData) {
                if (filter.accept(row)) {
                    displayableTableData.add(row);
                }
            }
        }
        sortTable();
    }


    /**
     * Sets filter.
     *
     * @param filter           to be set.
     * @param applyImmediately if <code>true</code>, table is filtered immediately
     */
    public void setFilter(Filter<TableRow> filter, boolean applyImmediately) {
        this.filter = filter;
        if (applyImmediately) {
            applyFilter();
        }
    }


    /**
     * Removes filter. Filter is identified by it's id.
     */
    public void dropFilter() {
        filter = DumbFilter.INSTANCE;
        applyFilter();
    }


    /**
     * Causes table data filtering and sorting. This method is usefull when
     * immediate operation was postponed in {@link #setFilter(ru.skipy.filtering.Filter, boolean)}
     * and/or {@link #setSortPreferences(ColumnSortPrefs[], boolean)}
     */
    public void filterAndSort() {
        applyFilter();
    }


    /**
     * Sends table change event to all registered listeners
     *
     * @param evt event to be send
     */
    private void fireTableChangedEvent(TableModelEvent evt) {
        for (TableModelListener listener : listeners) {
            listener.tableChanged(evt);
        }
    }


    /**
     * This is a trivial implementation of {@link Filter} interface.
     * Its' {@link #accept(ru.skipy.filtering.Filterable)} method always return
     * <code>true</code>
     *
     * @author Eugene Matyushkin
     * @version 1.0
     */
    private static class DumbFilter implements Filter<TableRow> {

        /**
         * Instance of the filter. Should be used when no filter set to avoid
         * multiple checks for <code>null</code>
         */
        private static final DumbFilter INSTANCE = new DumbFilter();


        /**
         * Accepts all rows
         *
         * @param row row to check
         *
         * @return always <code>true</code>
         */
        public boolean accept(TableRow row) {
            return true;
        }


        /**
         * No description available for this filter. It accepts all
         *
         * @return always <code>"true"</code> string
         */
        public String getHumanreadableRule() {
            return "true";
        }


        /**
         * Returns filter name
         *
         * @return filter name
         */
        public String getName() {
            return "DumbFilter";
        }


        /**
         * Returns filter description
         *
         * @return filter description
         */
        public String getDescription() {
            return "Accepts all Filterable elements";
        }

    }

}
